from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from six import PY2, text_type as str
from six.moves.urllib.parse import parse_qs, unquote, unquote_plus

from .mapping import Mapping


def _decode(o, errors='strict'):
    return o.decode('utf8', errors=errors) if isinstance(o, bytes) else o


def path_decode(bs):
    return _decode(unquote(bs.encode('ascii') if PY2 else bs))


class PathPart(str):
    """Represents a segment of a URL path.

    Attributes:
        params (Mapping): extra data attached to this segment
    """

    __slots__ = ['params']

    def __new__(cls, value, params=None):
        obj = super(PathPart, cls).__new__(cls, value)
        obj.params = params
        return obj

    def __repr__(self):
        return '%s(%r, params=%r)' % (self.__class__.__name__, str(self), self.params)


def extract_rfc2396_params(path):
    """This function implements parsing URL path parameters, per `section 3.3 of RFC2396`_.

    * path should be raw so we don't split or operate on a decoded character
    * output is decoded

    Example:

    >>> path = '/frisbee;color=red;size=small/logo;sponsor=w3c;color=black/image.jpg'
    >>> extract_rfc2396_params(path) == [
    ...     PathPart('frisbee', params={'color': ['red'], 'size': ['small']}),
    ...     PathPart('logo', params={'sponsor': ['w3c'], 'color': ['black']}),
    ...     PathPart('image.jpg', params={})
    ... ]
    True

    .. _Section 3.3 of RFC2396: https://tools.ietf.org/html/rfc3986#section-3.3
    """
    pathsegs = path.lstrip('/').split('/')
    segments_with_params = []
    for component in pathsegs:
        parts = component.split(';')
        params = Mapping()
        segment = path_decode(parts[0])
        for p in parts[1:]:
            if '=' in p:
                k, v = p.split('=', 1)
            else:
                k, v = p, ''
            params.add(path_decode(k), path_decode(v))
        segments_with_params.append(PathPart(segment, params))
    return segments_with_params


def split_path_no_params(path):
    """This splits a path into parts on "/" only (no split on ";" or ",").
    """
    return [PathPart(path_decode(s)) for s in path.lstrip('/').split('/')]


class Path(Mapping):
    """Represent the path of a resource.

    Attributes:
        raw: the unparsed form of the path - :class:`bytes`
        decoded: the decoded form of the path - :class:`str`
        parts: the segments of the path - :class:`list` of :class:`PathPart` objects
    """

    def __init__(self, raw, split_path=extract_rfc2396_params):
        self.raw = raw
        self.decoded = path_decode(raw)
        self.parts = split_path(raw)


class Querystring(Mapping):
    """Represent an HTTP querystring.

    Attributes:
        raw: the unparsed form of the querystring - :class:`bytes`
        decoded: the decoded form of the querystring - :class:`str`
    """

    def __init__(self, raw, errors='replace'):
        """Takes a string of type application/x-www-form-urlencoded.
        """
        # urllib needs bytestrings in py2 and unicode strings in py3
        raw_str = raw.encode('ascii') if PY2 else raw

        self.decoded = _decode(unquote_plus(raw_str), errors=errors)
        self.raw = raw

        common_kw = dict(keep_blank_values=True, strict_parsing=False)
        if PY2:
            # in python 2 parse_qs does its own unquote_plus'ing ...
            as_dict = parse_qs(raw_str, **common_kw)
            # ... but doesn't decode to unicode.
            for k, vals in list(as_dict.items()):
                as_dict[_decode(k, errors=errors)] = [
                    _decode(v, errors=errors) for v in vals
                ]
        else:
            # in python 3 parse_qs does the decoding
            as_dict = parse_qs(raw_str, errors=errors, **common_kw)

        Mapping.__init__(self, as_dict)



#!/usr/bin/env python
"""Benchmark try vs. except

without raise Exception: 0.94 seconds
with raise Exception: 9.26 seconds

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import time

if __name__ == '__main__':
    start = time.time()

    for i in range(10000000):
        try:
            raise Exception
            pass
        except Exception:
            pass

    end = time.time()
    print("%5.2f" % (end - start))



"""
The request processor dispatches requests to the filesystem, typecasts URL
path variables, loads the resource from the filesystem, and then renders and
encodes the resource (if it's dynamic).
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from copy import copy
import errno
import mimetypes
import os
import sys
from collections import defaultdict

from . import typecasting
from .dispatcher import DispatchStatus, HybridDispatcher, UserlandDispatcher
from .typecasting import defaults as default_typecasters
from ..resources import Resources
from ..http.resource import Static
from ..exceptions import ConfigurationError


default_indices = [
    'index.html', 'index.json', 'index',
    'index.html.spt', 'index.json.spt', 'index.spt',
]


class RequestProcessor(object):
    """A core request processor designed for integration into a host framework.

    The ``kwargs`` are for configuration, see :class:`DefaultConfiguration`
    for valid keys and default values.
    """

    def __init__(self, **kwargs):
        # Do some base-line configuration.
        # ================================
        # We want to do the following configuration of our Python environment
        # regardless of the user's configuration preferences

        # mimetypes
        aspens_mimetypes = os.path.join(os.path.dirname(__file__), 'mime.types')
        mimetypes.knownfiles += [aspens_mimetypes]
        # mimetypes.init is called below after the user has a turn.

        # XXX register codecs here


        # Configure from defaults and kwargs.
        # ===================================

        defaults = [(k, v) for k, v in DefaultConfiguration.__dict__.items() if k[0] != '_']
        for name, default in sorted(defaults):
            if name in kwargs:
                self.__dict__[name] = kwargs[name]
            else:
                self.__dict__[name] = copy(default)


        # Set some attributes.
        # ====================

        def safe_getcwd(errorstr):
            try:
                # If the working directory no longer exists, then the following
                # will raise OSError: [Errno 2] No such file or directory. I
                # swear I've seen this under supervisor, though I don't have
                # steps to reproduce. :-(  To get around this you specify a
                # www_root explicitly, or you can use supervisor's cwd
                # facility.

                return os.getcwd()
            except OSError as err:
                if err.errno != errno.ENOENT:
                    raise
                raise ConfigurationError(errorstr)

        # project root
        if self.project_root is not None:
            # canonicalize it
            if not os.path.isabs(self.project_root):
                cwd = safe_getcwd(
                    "Could not get a current working directory. You can specify "
                    "project_root in kwargs."
                )
                self.project_root = os.path.join(cwd, self.project_root)

            self.project_root = os.path.realpath(self.project_root)
            self.resource_directories.insert(0, self.project_root)

            # mime.types
            users_mimetypes = os.path.join(self.project_root, 'mime.types')
            mimetypes.knownfiles += [users_mimetypes]

            # PYTHONPATH
            sys.path.insert(0, self.project_root)

        # www_root
        if self.www_root is None:
            self.www_root = safe_getcwd(
                "Could not get a current working directory. You can specify "
                "www_root in kwargs."
            )

        self.www_root = os.path.realpath(self.www_root)
        self.resource_directories.insert(0, self.www_root)

        # kludge simplates -- should move out into a simplate plugin
        from ..simplates.renderers import factories
        from ..simplates.simplate import Simplate, SimplateDefaults
        Simplate.renderer_factories = factories(self)
        Simplate.default_renderers_by_media_type = defaultdict(lambda: self.renderer_default)
        Simplate.default_renderers_by_media_type[self.media_type_json] = 'json_dump'

        initial_context = { 'request_processor': self }
        Simplate.defaults = SimplateDefaults(
            Simplate.default_renderers_by_media_type,
            Simplate.renderer_factories,
            initial_context
        )

        # set up dynamic class mapping
        self.dynamic_classes_by_file_extension = dict(spt=Simplate)

        # create the dispatcher
        if self.dispatcher_class is None:
            self.dispatcher_class = (
                HybridDispatcher if self.changes_reload else UserlandDispatcher
            )
        self.dispatcher = self.dispatcher_class(
            self.www_root, self.is_dynamic, self.indices, self.typecasters,
            **kwargs.get('dispatcher_options', {})
        )
        self.dispatcher.build_dispatch_tree()

        # mime.types
        # ==========
        # It turns out that init'ing mimetypes is somewhat expensive. This is
        # significant in testing, though in dev/production you wouldn't notice.
        # In any case this means that if a devuser inits mimetypes themselves
        # then we won't do so again here, which is fine. Right?

        if not mimetypes.inited:
            mimetypes.init()

        # create the resources cache
        self.resources = Resources(self)


    def dispatch(self, path):
        """Call the dispatcher and inject the path variables into the given path object.

        Args:
            path (Path): the requested path, e.g. :obj:`'/foo'`

        Returns:
            A :class:`DispatchResult` object.
        """
        dispatch_result = self.dispatcher.dispatch(path.decoded, path.parts)
        if dispatch_result.wildcards:
            for k, v in dispatch_result.wildcards.items():
                path[k] = v
        return dispatch_result


    def process(self, path, querystring, accept_header, context):
        """Process a request.

        Args:
            path (Path): the requested path, e.g. :obj:`Path('/foo')`
            querystring (Querystring): the query parameters, e.g. :obj:`Querystring('?bar=baz')`
            accept_header (str): the value of the HTTP header ``Accept``
            context (dict): the context variables passed to dynamic resources

        Returns:
            A 3-tuple ``(dispatch_result, resource, output)``. The latter two are
            set to :obj:`None` if dispatching failed.

        """

        dispatch_result = self.dispatch(path)

        typecasting.apply_typecasters(self.typecasters, path, context)

        if dispatch_result.match and dispatch_result.status == DispatchStatus.okay:
            resource = self.resources.get(dispatch_result.match)
            context['querystring'] = querystring
            output = resource.render(context, dispatch_result, accept_header)
            if not isinstance(output.body, bytes):
                output.charset = self.encode_output_as
                output.body = output.body.encode(output.charset)
            return dispatch_result, resource, output

        return dispatch_result, None, None


    def is_dynamic(self, fspath):
        """Given a filesystem path, return a boolean.
        """
        return self.get_resource_class(fspath) is not Static


    def get_resource_class(self, fspath):
        """Given a filesystem path, return a resource class.
        """
        parts = fspath.split('.')
        extension = parts[-1] if len(parts) > 1 else None
        return self.dynamic_classes_by_file_extension.get(extension, Static)


    def guess_media_type(self, filename):
        """Guess the media type of a file by looking at its extension.

        This method is a small wrapper around :func:`mimetypes.guess_type`. It
        returns :attr:`~DefaultConfiguration.media_type_default` if the guessing
        fails.
        """
        media_type = mimetypes.guess_type(filename, strict=False)[0]
        if not media_type:
            media_type = self.media_type_default
        elif media_type == 'application/json':
            media_type = self.media_type_json
        return media_type


class DefaultConfiguration(object):
    """Default configuration values.
    """

    changes_reload = False
    """
    Reload files on every request if they've been modified. This can be costly,
    so it should be turned off in production.
    """

    charset_static = None
    """
    The charset of your static files. It ends up as a ``charset=`` parameter in
    ``Content-Type`` HTTP headers (if the framework on top of Aspen supports that).
    """

    dispatcher_class = None
    """
    The kind of dispatcher that will be used to route requests to files. By
    default :class:`UserlandDispatcher` is used, unless ``changes_reload`` is
    set to ``True``, then :class:`HybridDispatcher` is used.
    """

    encode_output_as = 'UTF-8'
    "The encoding to use for dynamically-generated output."

    indices = default_indices
    "List of file names that will be treated as directory indexes. The order matters."

    media_type_default = 'text/plain'
    "If the ``Content-Type`` of a response can't be determined, then this one is used."

    media_type_json = 'application/json'
    "The media type to use for the JSON format."

    project_root = None
    "The root directory of your project."

    renderer_default = 'stdlib_percent'
    "The default renderer for simplates."

    resource_directories = []
    """
    The list of directories containing resource files. Aspen will refuse to load
    any resource located outside these directories. The `www_root` directory is
    automatically added to this list. The `project_root` directory is added as
    well if it exists.
    """

    store_static_files_in_ram = False
    "If set to ``True``, store the contents of static files in RAM."

    typecasters = default_typecasters
    "See :mod:`aspen.request_processor.typecasting`."

    www_root = None
    """
    The root directory of your web app, containing the files it will serve.
    Defaults to the current directory.
    """



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import Renderer, Factory
from string import Template


class Renderer(Renderer):

    def compile(self, filepath, padded):
        return Template(padded)

    def render_content(self, context):
        return self.compiled.substitute(context)


class Factory(Factory):
    Renderer = Renderer



"""Importing this package will upgrade the `backslashreplace` error handling
scheme in the global `codecs` registry to support handling decoding errors as
well as encoding errors.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


from . import backcompat

backcompat.upgrade_backslashreplace()



"""
This module provides helpers for testing applications that use Aspen.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from contextlib import contextmanager
import os
import sys

from filesystem_tree import FilesystemTree

from .http.request import Path, Querystring
from .request_processor import RequestProcessor
from .request_processor.dispatcher import TestDispatcher


CWD = os.getcwd()


def teardown():
    """Standard teardown function.

    - reset the current working directory
    - remove FSFIX = %{tempdir}/fsfix
    - clear out sys.path_importer_cache

    """
    os.chdir(CWD)
    # Reset some process-global caches. Hrm ...
    sys.path_importer_cache = {} # see test_weird.py

teardown() # start clean


def resolve_want(available, want):
    if '.' in want:
        attr_path = want.split('.')
        base, attr_path = attr_path[0], attr_path[1:]
    else:
        base, attr_path = want, None
    try:
        out = available[base]
    except KeyError as e:
        raise KeyError('%r not found. available = %r' % (e, available))
    if attr_path:
        try:
            for name in attr_path:
                out = getattr(out, name)
        except AttributeError as e:
            raise AttributeError('%r not found. out = %r' % (e, out))
    return out


class FileSystem(object):
    __slots__ = ('project', 'www')

    def __init__(self):
        self.project = FilesystemTree()
        self.www = FilesystemTree()


class Harness(object):
    """A harness to be used in the Aspen test suite itself. Probably not useful to you.
    """

    def __init__(self):
        self.fs = FileSystem()
        self._request_processor = None

    def teardown(self):
        self.fs.www.remove()
        self.fs.project.remove()

    def hydrate_request_processor(self, **kwargs):
        if (self._request_processor is None) or kwargs:
            _kwargs = { 'www_root': self.fs.www.root
                      , 'project_root': self.fs.project.root
                      , 'dispatcher_class': TestDispatcher
                       }
            _kwargs.update(kwargs)
            self._request_processor = RequestProcessor(**_kwargs)
        return self._request_processor

    request_processor = property(hydrate_request_processor)

    def simple(self, contents='Greetings, program!', filepath='index.html.spt', uripath=None,
            querystring='', request_processor_configuration=None, **kw):
        """A helper to create a file and hit it through our machinery.
        """
        if filepath is not None:
            if isinstance(contents, tuple):
                contents, encoding = contents
            else:
                encoding = 'utf8'
            self.fs.www.mk((filepath, contents, True, encoding))
            if self._request_processor:
                # Rebuild the dispatch tree since we've just added a file
                self._request_processor.dispatcher.build_dispatch_tree()
        if request_processor_configuration is not None:
            self.hydrate_request_processor(**request_processor_configuration)

        if uripath is None:
            if filepath is None:
                uripath = '/'
            else:
                uripath = '/' + filepath
                if uripath.endswith('.spt'):
                    uripath = uripath[:-len('.spt')]
                for indexname in self.request_processor.indices:
                    if uripath.endswith(indexname):
                        uripath = uripath[:-len(indexname)]
                        break

        return self.hit(uripath, querystring, **kw)

    def hit(self, path, querystring='', accept_header=None, want=None, **context):
        path = context['path'] = Path(path)
        querystring = Querystring(querystring)
        dispatch_result, resource, output = self.request_processor.process(
            path, querystring, accept_header, context
        )
        if want is None:
            return output
        return resolve_want(locals(), want)


@contextmanager
def chdir(path):
    """A context manager that temporarily changes the working directory.
    """
    back_to = os.getcwd()
    os.chdir(path)
    try:
        yield back_to
    finally:
        os.chdir(back_to)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pytest import raises
from six import PY2


def test_with_non_ascii_bytes_and_without_encoding(harness):
    harness.fs.www.mk(('index.spt', """
        [------------------]
        text = u'א'
        [------------------] text/plain
        %(text)s
    """, 'utf8'))
    if PY2:
        with raises(UnicodeDecodeError):
            harness.hit('/')
    else:
        output = harness.hit('/')
        assert output.text.strip() == 'א'

def test_non_ascii_bytes_work_with_encoding(harness):
    expected = 'א'.encode('utf8')
    actual = harness.simple(("""
        # encoding=utf8
        [------------------]
        text = u'א'
        [------------------]
        %(text)s
    """, 'utf8')).body.strip()
    assert actual == expected

def test_the_exec_machinery_handles_two_encoding_lines_properly(harness):
    expected = 'א'.encode('utf8')
    actual = harness.simple(("""\
        # encoding=utf8
        # encoding=ascii
        [------------------]
        text = u'א'
        [------------------]
        %(text)s
    """, 'utf8')).body.strip()
    assert actual == expected



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os

from aspen.request_processor import RequestProcessor
from aspen.testing import chdir


def test_basic(harness):
    with chdir(harness.fs.www.root):
        rp = RequestProcessor()
        expected = os.getcwd()
        actual = rp.www_root
        assert actual == expected

def test_processor_can_process(harness):
    output = harness.simple('[---]\n[---]\nGreetings, program!', 'index.html.spt')
    assert output.text == 'Greetings, program!'

def test_user_can_influence_render_context(harness):
    assert harness.simple('[---]\n[---]\n%(foo)s', 'index.html.spt', foo='bar').text == 'bar'

def test_resources_can_import_from_project_root(harness):
    harness.fs.project.mk(('foo.py', 'bar = "baz"'))
    assert harness.simple( "from foo import bar\n[---]\n[---]\nGreetings, %(bar)s!"
                         , 'index.html.spt').text == "Greetings, baz!"



from os.path import dirname, join

try:
    import setuptools  # noqa
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()

from setuptools import find_packages, setup

version = open('version.txt').read()


classifiers = [ 'Development Status :: 4 - Beta'
              , 'Environment :: Console'
              , 'Intended Audience :: Developers'
              , 'License :: OSI Approved :: MIT License'
              , 'Natural Language :: English'
              , 'Operating System :: OS Independent'
              , 'Programming Language :: Python :: 2.7'
              , 'Programming Language :: Python :: 3.4'
              , 'Programming Language :: Python :: 3.5'
              , 'Programming Language :: Python :: 3.6'
              , 'Programming Language :: Python :: 3.7'
              , 'Programming Language :: Python :: Implementation :: CPython'
              , 'Topic :: Internet :: WWW/HTTP :: WSGI :: Application'
               ]

setup( author = 'Chad Whitacre et al.'
     , author_email = 'team@aspen.io'
     , classifiers = classifiers
     , description = 'A filesystem router for Python web frameworks'
     , long_description=open(join(dirname(__file__), 'README.rst')).read()
     , name = 'aspen'
     , packages = find_packages()
     , url = 'https://github.com/AspenWeb/aspen.py'
     , version = version
     , zip_safe = False
     , package_data = {'aspen': ['request_processor/mime.types']}
     , install_requires = open('requirements.txt').read()
     , tests_require = open('requirements_tests.txt').read()
      )




REPR_TEMPLATE = """
def __repr__(self):
    return '{}({})' % ({})
"""

def auto_repr(cls):
    """Generates a `__repr__` method for the given class, using `__slots__`.

    >>> @auto_repr
    ... class Foo(object):
    ...     __slots__ = ['bar']
    ...     def __init__(self, bar):
    ...         self.bar = bar

    >>> Foo(0)
    Foo(bar=0)
    """
    slots = cls.__slots__
    repr_slots = ', '.join('{}=%r'.format(attr) for attr in slots)
    repr_values = ', '.join('self.' + attr for attr in slots)
    repr_def = REPR_TEMPLATE.format(cls.__name__, repr_slots, repr_values)
    namespace = dict(__name__='auto_repr_%s' % cls.__name__)
    exec(repr_def, namespace)
    cls.__repr__ = namespace['__repr__']
    try:
        cls.__repr__._source = repr_def
    except AttributeError:
        pass
    return cls


class Constant(object):
    """A simple class that creates lightweight constants.

    >>> c = Constant('OK')
    >>> c
    OK
    >>> c.name
    'OK'
    >>> c == c
    True

    >>> c.name = 'KO'
    Traceback (most recent call last):
    ...
    AttributeError: constants cannot be modified

    >>> c2 = Constant('OK')
    >>> c2 == c
    False
    """

    __slots__ = 'name'

    def __init__(self, name):
        object.__setattr__(self, 'name', name)

    def __repr__(self):
        return self.name

    def __setattr__(self, name, value):
        raise AttributeError("constants cannot be modified")



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from . import Renderer, Factory


class Renderer(Renderer):

    def render_content(self, context):
        return self.compiled.format(**context)


class Factory(Factory):
    Renderer = Renderer



#!/usr/bin/env python
"""Benchmark assert

without assert: 0.79 seconds
with assert: 0.95 seconds

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import time

if __name__ == '__main__':
    start = time.time()

    for i in range(10000000):
        #pass
        assert 1 == 1

    end = time.time()
    print("%5.2f" % (end - start))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import stat


class Entry(object):
    """An entry in a resource cache.
    """
    __slots__ = ('fspath', 'mtime', 'resource')

    def __init__(self, fspath, mtime, resource):
        #: The filesystem path [string]
        self.fspath = fspath
        #: The timestamp of the last change [int]
        self.mtime = mtime
        #: The loaded resource [Static or Dynamic]
        self.resource = resource


class Resources(object):
    """This class implements loading resources, and caching them.
    """

    __slots__ = ('request_processor', 'cache')

    def __init__(self, request_processor):
        self.request_processor = request_processor
        self.cache = {}

    def get(self, fspath):
        """Return a resource object, with caching.
        """

        # Get a cache Entry object.
        entry = self.cache.get(fspath)

        # Process the resource.
        if not entry or self.request_processor.changes_reload:
            mtime = os.stat(fspath)[stat.ST_MTIME]
            if getattr(entry, 'mtime', None) != mtime:  # cache miss
                resource = self.load(fspath)
                entry = self.cache[fspath] = Entry(fspath, mtime, resource)

        return entry.resource

    def load(self, fspath):
        """Create and return a resource object, without caching.
        """
        Class = self.request_processor.get_resource_class(fspath)
        return Class(self.request_processor, fspath)
